[![Warp Logo](https://www.scaledynamics.io/assets/images/banner/warp-npm.png)](https://scaledynamics.io)

# Warp

_Warp_ is an API allowing easy parallel execution of JavaScript code, at the
same physical time as the main JavaScript thread and as other threads.

```javascript
// runs `fnct` in parallel
warp.call(fnct, <args>, cb);
```

Discover Warp in this short YouTube video:

[![Warp Principle](https://raw.githubusercontent.com/ScaleDynamics/warp/d600e20784b2253b846f0301adb2f31aca328256/doc/images/warp-principle-video.png)](https://www.youtube.com/watch?v=d54AOpRgfg0)

## Installing

_Warp_ module is an API and is not providing support for execution. You will need
to acquire a [**Warp runtime**](https://redirect.scaledynamics.io/warp/runtime) to use it.

To add _Warp_ API to your project:

```
$ npm install warp
```

## How to use it

_Warp_ offers two main functions starting a [_warp_ task](#warp-task): `warp.call` and `warp.callAsPromise`.
We say that they perform a _warp_ call. By extension, we also say that code
_warps_ a function or that the function is _warped_.

Here is a very simple illustrative code sample where a CPU intensive function (`computePrimeFactorization`)
is executed in parallel by a _warp_ call keeping the event loop free for any other processing. If you
want to see complete code, browse the [repository](https://github.com/ScaleDynamics/warp/tree/d600e20784b2253b846f0301adb2f31aca328256/example).

The first example uses a callback-based function, whereas the second uses a Promise-based function.

### With callback

``` javascript
const warp = require('warp');

function computePrimeFactorization(n, done) {
  // Compute `f`
  done(f);
}

warp.call(computePrimeFactorization, 9007199254740991,
    (a) => console.log(`Prime factors of ${a.value} are ${a.factors}`));
```

Prints out
> Prime factors of 9007199254740991 are 6361,69431,20394401

### With promise

``` javascript
const warp = require('warp');

function computePrimeFactorization(n) {
  // Compute `f`
  return f;
}

let a = await warp.callAsPromise(computePrimeFactorization, 9007199254740991);
console.log(`Prime factors of ${a.value} are ${a.factors}`);
```

Prints out
> Prime factors of 9007199254740991 are 6361,69431,20394401

## How it works

![Warp Principle](https://raw.githubusercontent.com/ScaleDynamics/warp/d600e20784b2253b846f0301adb2f31aca328256/doc/images/warp-principle.png)

When you _warp_ call a function, the [execution context](#execution-context) (i.e. everything
reachable from the function scope) is captured, then this context and the _warped_ function are
sent to a distinct and [parallel _warp_ thread](#parallel-thread) (see label `A` in the drawing).

Once the [_warp_ task](#warp-task) is finished, the parallel [context](#execution-context) is
captured and sent back to the main JavaScript thread  (see label `B` in the drawing). Then,
in an event task of the main thread, the context is updated, and:

- If the _warp_ call is `call`, the callback is executed in the main thread, with the arguments
given to the callback in the parallel _warp_ thread.
- If the _warp_ call is `callAsPromise`, the Promise returned by `callAsPromise` is resolved or
rejected with the value or the exception obtained at the end of the parallel _warp_ thread.

### Scope

The _warped_ functions can read and modify anything reachable from their
[scope](https://developer.mozilla.org/en-US/docs/Glossary/Scope).

In the following sample, the variable `o` is defined as a free variable (outside of the function).
The _warped_ function can read it, and also modify the object. Modifications made to `o` in the parallel
_warp_ thread are applied and visible in the main thread.

``` javascript
const warp = require('warp');

let o = {};
function computePrimeFactorization(n) {
  // Compute `f`
  o.result = f;
}

await warp.callAsPromise(computePrimeFactorization, 9007199254740991);
console.log(`Prime factors of ${o.value} are ${o.factors}`);
```

Prints out
> Prime factors of 9007199254740991 are 6361,69431,20394401

### Shared resources and concurrency

_Warp_ offers no new synchronization or semaphore mechanism. If you need to manage concurrent access
to a value, you'll have to keep the modifying code in the main thread.

In the following example, we compute several prime factorizations in parallel.

``` javascript
const warp = require('warp');

let primeFactorizations = [];

function computePrimeFactorization(index, n) {
  // Compute `f`
  primeFactorizations[index] = f;
}

let promises = []
for(let i = 0; i < 8; i++) {
  promises.push(warp.callAsPromise(computePrimeFactorization, i, 9007174991 + i));
}

await Promise.all(promises);

console.log(`Prime factorizations: %j`, primeFactorizations);
```

Prints out
> Prime factorizations: [{"value":9007174991,"factors":[17,131,4044533]},{"value":9007174997,"factors":{"0":59,"1":109,"2":1400587,"value":9007174993,"factors":[12071,746183]}},{"0":59,"1":109,"2":1400587,"value":9007174993,"factors":[12071,746183]},{"0":2,"1":2,"2":2251793749,"value":9007174994,"factors":[2,37,53,2296577]},{"value":9007174995,"factors":[3,5,7,13,41,227,709]},{"value":9007174997,"factors":{"0":59,"1":109,"2":1400587,"value":9007174993,"factors":[12071,746183]}},{"value":9007174997,"factors":{"0":59,"1":109,"2":1400587,"value":9007174993,"factors":[12071,746183]}},{"value":9007174998,"factors":[2,3,3,3,19,43,204161]}]

But you need to be carefull about parallel access to shared variables or properties.

For instance, the following example shows bad usage of a shared value causing indetermism. There, you
don't know in which order the additions will be made, but it is guaranteed that each `fnct` sees
the initial value of `v` as `4`. This means you don't know what the final value of `v` will be, you
just know it will be either `7` or `8`.

``` javascript
const warp = require('warp');

let v = 4;
function fnct(num) {
  let inc = num;
  v += inc;
  return;
}

let promises = []
promises.push(warp.callAsPromise(fnct, 3));
promises.push(warp.callAsPromise(fnct, 4));

await Promise.all(promises);

console.log(`Value: ${v}`);
```

To obtain the correct behavior in this example, you'll need to apply the changes to the shared
variable in the main thread by using the `.then` of the promise.

``` javascript
const warp = require('warp');

let v = 4;
function fnct(num) {
  let inc = num;
  return inc;
}

let promises = []
promises.push(warp.callAsPromise(fnct, 3).then((inc) => v += inc));
promises.push(warp.callAsPromise(fnct, 4).then((inc) => v += inc));

await Promise.all(promises);

console.log(`Value: ${v}`);
```

Prints out
> Value: 11

This way it'll end up with `11` as the final value.

### Use transferable objects

``` javascript
let arrayBuffer = new ArrayBuffer(8);
warp.markTransferable(arrayBuffer);
```

As for the Worker.postMessage method, some objects can be identified as transferable. You must pass all your Transferable objects to _warp_ markTransferable function to transfer these objects. In that case, it becomes unusable (neutered) in the main thread and becomes available only to the parallel thread it was sent to.
Transferable objects are instances of classes like ArrayBuffer, MessagePort, OffSceenCanvas or ImageBitmap objects.
ArrayBuffer can be transferred conventionally but this type of transfer is faster.

# Definitions

## <a name="parallel-thread"></a>Parallel thread

A **parallel thread** is a JavaScript execution thread distinct from the main JavaScript execution thread. The
main thread and some parallel _warp_ threads are running at the same physical time. _Warp_ threads don't necessarily belong to the same process. It can come from the fork of a new process or even in a remote machine.

## <a name="warp-task"></a>Warp task

A **warp task** is started by a _warp_ call (a call to `warp.call` or `warp.callAsPromise`).
The task is executed in a [parallel thread](#parallel-thread). It starts by executing the given
function and continues until the end of the task is detected.

The end of the task is detected when:
- an exception is thrown by the task

**or**

- If the task was started by `warp.callAsPromise`:   
  - the function given to `callAsPromise` has returned   
  - **And**, if its returned value is a Promise, it is
  [fulfilled or rejected](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise#Description).

- If the task was started by `warp.call`   
  - The callback is called,   
  - **And** the function given to `call` has returned.   

Note that this doesn't mean that `return` was executed, it means that the execution has
returned to its caller. For example in the following code the end is detected during the line
`a = await 3` when `await` yields and `a` is still `2`.

``` javascript
let a = 0;
async function fnct(done) {
  a = 1;
  done();
  a = 2;
  a = await 3; // End is here at `await` before assignation of `a`
  return;
}
```

**Warning:** If the task were to attempt to execute code after the end of execution were detected
(e.g. with a `setImmediate`, `setTimeout`, ...), then the behavior would be unspecified. This
means that if the function given to `call` is starting any asynchronous processing, calling
the `callback` must be the very last thing done by the task.

## <a name="execution-context"></a>Execution context

An **execution context** is referring to everything accessible from the
[scope](https://developer.mozilla.org/en-US/docs/Glossary/Scope) of the considered
function at a given point of the execution. This includes   
- parameters of the call,   
- local, free and global variables accessed by the function called or by all functions potentially
  called directly or indirectly,   
- any reachable objects (objects, arrays, functions, symbols) from the above defined parameters/variables.   
